program HeapDemo;

{$APPTYPE CONSOLE}

uses
  Windows,
  SysUtils,
  TlHelp32;

procedure QueryProcessHeap1;
  var
    hSnapShot: THandle;
    HeapList: THeapList32;
    HeapEntry: THeapEntry32;
    Start: DWORD;
  begin
    Start := GetTickCount;
    Writeln('Heap info:');
    Writeln;
    hSnapShot := CreateToolhelp32Snapshot(TH32CS_SNAPHEAPLIST, 0);
    try
      HeapList.dwSize := SizeOf(THeapList32);
      if Heap32ListFirst(hSnapShot, HeapList) then
      repeat
        HeapEntry.dwSize := SizeOf(THeapEntry32);
        if Heap32First(HeapEntry, GetCurrentProcessId, HeapList.th32HeapID) then
        repeat
          Writeln(Format('Heap addr: 0x%p, size: %d',
            [Pointer(HeapEntry.dwAddress), HeapEntry.dwBlockSize]));
        until not Heap32Next(HeapEntry);
      until not Heap32ListNext(hSnapShot, HeapList);
    finally
      CloseHandle(hSnapShot);
    end;
    Writeln;
    Writeln('DONE. Time elapsed: ', GetTickCount - Start);
  end;

const
  ntdll = 'ntdll.dll';
  RTL_QUERY_PROCESS_HEAP_SUMMARY = $00000004;
  RTL_QUERY_PROCESS_HEAP_TAGS = $00000008;
  RTL_QUERY_PROCESS_HEAP_ENTRIES = $00000010;
  RTL_HEAP_SEGMENT = 2;

type
  PVOID = Pointer;
  HANDLE = THandle;
  NTSTATUS = DWORD;

  _CLIENT_ID = record
    UniqueProcess: HANDLE;
    UniqueThread: HANDLE;
  end;
  CLIENT_ID = _CLIENT_ID;
  PCLIENT_ID = ^CLIENT_ID;
  TClientID = CLIENT_ID;
  PClientID = ^TClientID;

  KAFFINITY = ULONG;
  PKAFFINITY = ^KAFFINITY;
  KPRIORITY = DWORD;

  _THREAD_BASIC_INFORMATION = record // Information Class 0
    ExitStatus: NTSTATUS;
    TebBaseAddress: Pointer;
    ClientId: CLIENT_ID;
    AffinityMask: KAFFINITY;
    Priority: KPRIORITY;
    BasePriority: KPRIORITY;
  end;
  THREAD_BASIC_INFORMATION = _THREAD_BASIC_INFORMATION;
  PTHREAD_BASIC_INFORMATION = ^THREAD_BASIC_INFORMATION;
  TThreadBasicInformation = THREAD_BASIC_INFORMATION;
  PThreadBasicInformation = ^TThreadBasicInformation;

  USHORT = Word;

  TSettableAndTag = record
    Settable: ULONG;
    Tag: ULONG;
  end;

  TCommittedSizeAndBlock = record
    CommittedSize: ULONG;
    FirstBlock: Pointer;
  end;

  THeapEntryUnion = record
  case Integer of
    0: (s1: TSettableAndTag);
    1: (s2: TCommittedSizeAndBlock);
  end;

  _RTL_HEAP_ENTRY = record
    Size: ULONG;
    Flags: USHORT;
    AllocatorBackTraceIndex: USHORT;
    u: THeapEntryUnion;
  end;
  RTL_HEAP_ENTRY = _RTL_HEAP_ENTRY;
  PRTL_HEAP_ENTRY = ^RTL_HEAP_ENTRY;
  TRtrHeapEntry = RTL_HEAP_ENTRY;
  PRtrHeapEntry = PRTL_HEAP_ENTRY;
  
//struct rtl_heap_tag
//{
//    uint32_t NumberOfAllocations;
//    uint32_t NumberOfFrees;
//    size_t BytesAllocated;
//    uint16_t TagIndex;
//    uint16_t CreatorBackTraceIndex;
//    wchar_t TagName[ 24 ];
//} ;


  _RTL_HEAP_INFORMATION = record
    BaseAddress: Pointer;
    Flags: ULONG;
    EntryOverhead: USHORT;
    CreatorBackTraceIndex: USHORT;
    BytesAllocated: ULONG;
    BytesCommitted: ULONG;
    NumberOfTags: ULONG;
    NumberOfEntries: ULONG;
    NumberOfPseudoTags: ULONG;
    PseudoTagGranularity: ULONG;
    Reserved: array [0..4] of ULONG;
    Tags: Pointer;
    Entries: PRTL_HEAP_ENTRY;
  end;
  RTL_HEAP_INFORMATION = _RTL_HEAP_INFORMATION;
  PRTL_HEAP_INFORMATION = ^RTL_HEAP_INFORMATION;
  TRtlHeapInformation = RTL_HEAP_INFORMATION;
  PRtlHeapInformation = ^TRtlHeapInformation;

  _RTL_PROCESS_HEAPS = record
    NumberOfHeaps: ULONG;
    Heaps: array [0..0] of RTL_HEAP_INFORMATION;
  end;
  RTL_PROCESS_HEAPS = _RTL_PROCESS_HEAPS;
  PRTL_PROCESS_HEAPS = ^RTL_PROCESS_HEAPS;
  TRtlProcessHeaps = RTL_PROCESS_HEAPS;
  PRtlProcessHeaps = ^TRtlProcessHeaps;

  _RTL_PROCESS_VERIFIER_OPTIONS = record
    SizeStruct: ULONG;
    Option: ULONG;
    OptionData: UCHAR;
  end;
  RTL_PROCESS_VERIFIER_OPTIONS = _RTL_PROCESS_VERIFIER_OPTIONS;
  PRTL_PROCESS_VERIFIER_OPTIONS = ^RTL_PROCESS_VERIFIER_OPTIONS;

  _RTL_DEBUG_INFORMATION = record
    SectionHandleClient: HANDLE;
    ViewBaseClient: PVOID;
    ViewBaseTarget: PVOID;
    ViewBaseDelta: ULONG;
    EventPairClient: HANDLE;
    EventPairTarget: PVOID;
    TargetProcessId: HANDLE;
    TargetThreadHandle: HANDLE;
    Flags: ULONG;
    OffsetFree: ULONG;
    CommitSize: ULONG;
    ViewSize: ULONG;
    {
    union
    {
        PRTL_PROCESS_MODULES Modules;
        PRTL_PROCESS_MODULE_INFORMATION_EX ModulesEx;
    }
    Modules: PVOID;
    BackTraces: PVOID; // PRTL_PROCESS_BACKTRACES
    Heaps: PRTL_PROCESS_HEAPS;
    Locks: PVOID; // PRTL_PROCESS_LOCKS
    SpecificHeap: HANDLE;
    TargetProcessHandle: HANDLE;
    VerifierOptions: RTL_PROCESS_VERIFIER_OPTIONS;
    ProcessHeap: HANDLE;
    CriticalSectionHandle: HANDLE;
    CriticalSectionOwnerThread: HANDLE;
    Reserved: array [0..3] of PVOID;
  end;
  RTL_DEBUG_INFORMATION = _RTL_DEBUG_INFORMATION;
  PRTL_DEBUG_INFORMATION = ^RTL_DEBUG_INFORMATION;
  TRtlDebugInformation = RTL_DEBUG_INFORMATION;
  PRtlDebugInformation = ^TRtlDebugInformation;

  function RtlCreateQueryDebugBuffer(
    Size: ULONG; EventPair: BOOLEAN): PRtlDebugInformation; stdcall;
    external ntdll;

  function RtlQueryProcessDebugInformation(
    ProcessId: ULONG; DebugInfoClassMask: ULONG;
    DebugBuffer: PRtlDebugInformation): NTSTATUS; stdcall; external ntdll;

  function RtlDestroyQueryDebugBuffer(
    DebugBuffer: PRtlDebugInformation): NTSTATUS; stdcall; external ntdll;

  procedure QueryProcessHeap2;
  var
    I, A: Integer;
    pDbgBuffer: PRtlDebugInformation;
    pHeapInformation: PRtlHeapInformation;
    pHeapEntry: PRtrHeapEntry;
    dwHeapStartAddr, dwAddr, dwLastSize: DWORD;
    hit_seg_count: Integer;
    Start: DWORD;
  begin
    Start := GetTickCount;
    Writeln('Heap info:');
    Writeln;
    pDbgBuffer := RtlCreateQueryDebugBuffer(0, False);
    if pDbgBuffer <> nil then
    try
      //      
      if RtlQueryProcessDebugInformation(GetCurrentProcessId,
        RTL_QUERY_PROCESS_HEAP_SUMMARY or RTL_QUERY_PROCESS_HEAP_ENTRIES,
        pDbgBuffer) = 0 then
      begin
        //      
        pHeapInformation := @pDbgBuffer^.Heaps^.Heaps[0];
        //    ...
        for I := 0 to pDbgBuffer^.Heaps^.NumberOfHeaps - 1 do
        begin
          //    
          pHeapEntry := pHeapInformation^.Entries;
          dwAddr := DWORD(pHeapEntry^.u.s2.FirstBlock) +
            pHeapInformation^.EntryOverhead;
          dwLastSize := 0;
          for A := 0 to pHeapInformation^.NumberOfEntries - 1 do
          begin
            dwHeapStartAddr := dwAddr;
            hit_seg_count := 0;
            while (pHeapEntry^.Flags and RTL_HEAP_SEGMENT) = RTL_HEAP_SEGMENT do
            begin
              //     RTL_HEAP_SEGMENT,
              //    
              dwAddr := DWORD(pHeapEntry^.u.s2.FirstBlock) +
                pHeapInformation^.EntryOverhead;
              Inc(pHeapEntry);
              Inc(hit_seg_count);
              //     
              if A + hit_seg_count >=
                Integer(pHeapInformation^.NumberOfEntries - 1) then
                Continue;
            end;
            //       ,     ,
            //    +   
            if hit_seg_count = 0 then
              Inc(dwAddr, dwLastSize);
            Writeln(Format('Heap addr: 0x%p, size: %d',
              [Pointer(dwHeapStartAddr), dwAddr - dwHeapStartAddr]));
           //    
           dwLastSize := pHeapEntry^.Size;
           //    
           Inc(pHeapEntry);
          end;
          //    
          Inc(pHeapInformation);
        end;
      end;
    finally
      RtlDestroyQueryDebugBuffer(pDbgBuffer);
    end;
    Writeln;
    Writeln('DONE. Time elapsed: ', GetTickCount - Start);
  end;

begin
  QueryProcessHeap1;
  QueryProcessHeap2;
  Readln;
end.
 